@@ -45,13 +45,15 @@ import (
 	"github.com/prometheus/prometheus/discovery/marathon"
 	"github.com/prometheus/prometheus/discovery/moby"
 	"github.com/prometheus/prometheus/discovery/openstack"
+	"github.com/prometheus/prometheus/discovery/puppetdb"
 	"github.com/prometheus/prometheus/discovery/scaleway"
 	"github.com/prometheus/prometheus/discovery/targetgroup"
 	"github.com/prometheus/prometheus/discovery/triton"
+	"github.com/prometheus/prometheus/discovery/uyuni"
 	"github.com/prometheus/prometheus/discovery/xds"
 	"github.com/prometheus/prometheus/discovery/zookeeper"
-	"github.com/prometheus/prometheus/pkg/labels"
-	"github.com/prometheus/prometheus/pkg/relabel"
+	"github.com/prometheus/prometheus/model/labels"
+	"github.com/prometheus/prometheus/model/relabel"
 )
 
 func mustParseURL(u string) *config.URL {
@@ -101,6 +103,10 @@ var expectedConf = &Config{
 					ClientID:     "123",
 					ClientSecret: "456",
 					TokenURL:     "http://remote1/auth",
+					TLSConfig: config.TLSConfig{
+						CertFile: filepath.FromSlash("testdata/valid_cert_file"),
+						KeyFile:  filepath.FromSlash("testdata/valid_key_file"),
+					},
 				},
 				FollowRedirects: true,
 			},
@@ -221,7 +227,6 @@ var expectedConf = &Config{
 			},
 		},
 		{
-
 			JobName: "service-x",
 
 			HonorTimestamps: true,
@@ -563,6 +568,7 @@ var expectedConf = &Config{
 					AuthenticationMethod: "OAuth",
 					RefreshInterval:      model.Duration(5 * time.Minute),
 					Port:                 9100,
+					HTTPClientConfig:     config.DefaultHTTPClientConfig,
 				},
 			},
 		},
@@ -777,17 +783,48 @@ var expectedConf = &Config{
 			Scheme:           DefaultScrapeConfig.Scheme,
 			HTTPClientConfig: config.DefaultHTTPClientConfig,
 
-			ServiceDiscoveryConfigs: discovery.Configs{&openstack.SDConfig{
-				Role:            "instance",
-				Region:          "RegionOne",
-				Port:            80,
-				Availability:    "public",
-				RefreshInterval: model.Duration(60 * time.Second),
-				TLSConfig: config.TLSConfig{
-					CAFile:   "testdata/valid_ca_file",
-					CertFile: "testdata/valid_cert_file",
-					KeyFile:  "testdata/valid_key_file",
-				}},
+			ServiceDiscoveryConfigs: discovery.Configs{
+				&openstack.SDConfig{
+					Role:            "instance",
+					Region:          "RegionOne",
+					Port:            80,
+					Availability:    "public",
+					RefreshInterval: model.Duration(60 * time.Second),
+					TLSConfig: config.TLSConfig{
+						CAFile:   "testdata/valid_ca_file",
+						CertFile: "testdata/valid_cert_file",
+						KeyFile:  "testdata/valid_key_file",
+					},
+				},
+			},
+		},
+		{
+			JobName: "service-puppetdb",
+
+			HonorTimestamps: true,
+			ScrapeInterval:  model.Duration(15 * time.Second),
+			ScrapeTimeout:   DefaultGlobalConfig.ScrapeTimeout,
+
+			MetricsPath:      DefaultScrapeConfig.MetricsPath,
+			Scheme:           DefaultScrapeConfig.Scheme,
+			HTTPClientConfig: config.DefaultHTTPClientConfig,
+
+			ServiceDiscoveryConfigs: discovery.Configs{
+				&puppetdb.SDConfig{
+					URL:               "https://puppetserver/",
+					Query:             "resources { type = \"Package\" and title = \"httpd\" }",
+					IncludeParameters: true,
+					Port:              80,
+					RefreshInterval:   model.Duration(60 * time.Second),
+					HTTPClientConfig: config.HTTPClientConfig{
+						FollowRedirects: true,
+						TLSConfig: config.TLSConfig{
+							CAFile:   "testdata/valid_ca_file",
+							CertFile: "testdata/valid_cert_file",
+							KeyFile:  "testdata/valid_key_file",
+						},
+					},
+				},
 			},
 		},
 		{
@@ -905,6 +942,26 @@ var expectedConf = &Config{
 				},
 			},
 		},
+		{
+			JobName: "uyuni",
+
+			HonorTimestamps:  true,
+			ScrapeInterval:   model.Duration(15 * time.Second),
+			ScrapeTimeout:    DefaultGlobalConfig.ScrapeTimeout,
+			HTTPClientConfig: config.HTTPClientConfig{FollowRedirects: true},
+			MetricsPath:      DefaultScrapeConfig.MetricsPath,
+			Scheme:           DefaultScrapeConfig.Scheme,
+			ServiceDiscoveryConfigs: discovery.Configs{
+				&uyuni.SDConfig{
+					Server:          "https://localhost:1234",
+					Username:        "gopher",
+					Password:        "hole",
+					Entitlement:     "monitoring_entitled",
+					Separator:       ",",
+					RefreshInterval: model.Duration(60 * time.Second),
+				},
+			},
+		},
 	},
 	AlertingConfig: AlertingConfig{
 		AlertmanagerConfigs: []*AlertmanagerConfig{
@@ -931,7 +988,7 @@ var expectedConf = &Config{
 }
 
 func TestYAMLRoundtrip(t *testing.T) {
-	want, err := LoadFile("testdata/roundtrip.good.yml", false, log.NewNopLogger())
+	want, err := LoadFile("testdata/roundtrip.good.yml", false, false, log.NewNopLogger())
 	require.NoError(t, err)
 
 	out, err := yaml.Marshal(want)
@@ -944,7 +1001,7 @@ func TestYAMLRoundtrip(t *testing.T) {
 }
 
 func TestRemoteWriteRetryOnRateLimit(t *testing.T) {
-	want, err := LoadFile("testdata/remote_write_retry_on_rate_limit.good.yml", false, log.NewNopLogger())
+	want, err := LoadFile("testdata/remote_write_retry_on_rate_limit.good.yml", false, false, log.NewNopLogger())
 	require.NoError(t, err)
 
 	out, err := yaml.Marshal(want)
@@ -960,16 +1017,16 @@ func TestRemoteWriteRetryOnRateLimit(t *testing.T) {
 func TestLoadConfig(t *testing.T) {
 	// Parse a valid file that sets a global scrape timeout. This tests whether parsing
 	// an overwritten default field in the global config permanently changes the default.
-	_, err := LoadFile("testdata/global_timeout.good.yml", false, log.NewNopLogger())
+	_, err := LoadFile("testdata/global_timeout.good.yml", false, false, log.NewNopLogger())
 	require.NoError(t, err)
 
-	c, err := LoadFile("testdata/conf.good.yml", false, log.NewNopLogger())
+	c, err := LoadFile("testdata/conf.good.yml", false, false, log.NewNopLogger())
 	require.NoError(t, err)
 	require.Equal(t, expectedConf, c)
 }
 
 func TestScrapeIntervalLarger(t *testing.T) {
-	c, err := LoadFile("testdata/scrape_interval_larger.good.yml", false, log.NewNopLogger())
+	c, err := LoadFile("testdata/scrape_interval_larger.good.yml", false, false, log.NewNopLogger())
 	require.NoError(t, err)
 	require.Equal(t, 1, len(c.ScrapeConfigs))
 	for _, sc := range c.ScrapeConfigs {
@@ -979,7 +1036,7 @@ func TestScrapeIntervalLarger(t *testing.T) {
 
 // YAML marshaling must not reveal authentication credentials.
 func TestElideSecrets(t *testing.T) {
-	c, err := LoadFile("testdata/conf.good.yml", false, log.NewNopLogger())
+	c, err := LoadFile("testdata/conf.good.yml", false, false, log.NewNopLogger())
 	require.NoError(t, err)
 
 	secretRe := regexp.MustCompile(`\\u003csecret\\u003e|<secret>`)
@@ -989,38 +1046,38 @@ func TestElideSecrets(t *testing.T) {
 	yamlConfig := string(config)
 
 	matches := secretRe.FindAllStringIndex(yamlConfig, -1)
-	require.Equal(t, 15, len(matches), "wrong number of secret matches found")
+	require.Equal(t, 16, len(matches), "wrong number of secret matches found")
 	require.NotContains(t, yamlConfig, "mysecret",
 		"yaml marshal reveals authentication credentials.")
 }
 
 func TestLoadConfigRuleFilesAbsolutePath(t *testing.T) {
 	// Parse a valid file that sets a rule files with an absolute path
-	c, err := LoadFile(ruleFilesConfigFile, false, log.NewNopLogger())
+	c, err := LoadFile(ruleFilesConfigFile, false, false, log.NewNopLogger())
 	require.NoError(t, err)
 	require.Equal(t, ruleFilesExpectedConf, c)
 }
 
 func TestKubernetesEmptyAPIServer(t *testing.T) {
-	_, err := LoadFile("testdata/kubernetes_empty_apiserver.good.yml", false, log.NewNopLogger())
+	_, err := LoadFile("testdata/kubernetes_empty_apiserver.good.yml", false, false, log.NewNopLogger())
 	require.NoError(t, err)
 }
 
 func TestKubernetesWithKubeConfig(t *testing.T) {
-	_, err := LoadFile("testdata/kubernetes_kubeconfig_without_apiserver.good.yml", false, log.NewNopLogger())
+	_, err := LoadFile("testdata/kubernetes_kubeconfig_without_apiserver.good.yml", false, false, log.NewNopLogger())
 	require.NoError(t, err)
 }
 
 func TestKubernetesSelectors(t *testing.T) {
-	_, err := LoadFile("testdata/kubernetes_selectors_endpoints.good.yml", false, log.NewNopLogger())
+	_, err := LoadFile("testdata/kubernetes_selectors_endpoints.good.yml", false, false, log.NewNopLogger())
 	require.NoError(t, err)
-	_, err = LoadFile("testdata/kubernetes_selectors_node.good.yml", false, log.NewNopLogger())
+	_, err = LoadFile("testdata/kubernetes_selectors_node.good.yml", false, false, log.NewNopLogger())
 	require.NoError(t, err)
-	_, err = LoadFile("testdata/kubernetes_selectors_ingress.good.yml", false, log.NewNopLogger())
+	_, err = LoadFile("testdata/kubernetes_selectors_ingress.good.yml", false, false, log.NewNopLogger())
 	require.NoError(t, err)
-	_, err = LoadFile("testdata/kubernetes_selectors_pod.good.yml", false, log.NewNopLogger())
+	_, err = LoadFile("testdata/kubernetes_selectors_pod.good.yml", false, false, log.NewNopLogger())
 	require.NoError(t, err)
-	_, err = LoadFile("testdata/kubernetes_selectors_service.good.yml", false, log.NewNopLogger())
+	_, err = LoadFile("testdata/kubernetes_selectors_service.good.yml", false, false, log.NewNopLogger())
 	require.NoError(t, err)
 }
 
@@ -1031,170 +1088,224 @@ var expectedErrors = []struct {
 	{
 		filename: "jobname.bad.yml",
 		errMsg:   `job_name is empty`,
-	}, {
+	},
+	{
 		filename: "jobname_dup.bad.yml",
 		errMsg:   `found multiple scrape configs with job name "prometheus"`,
-	}, {
+	},
+	{
 		filename: "scrape_interval.bad.yml",
 		errMsg:   `scrape timeout greater than scrape interval`,
-	}, {
+	},
+	{
 		filename: "labelname.bad.yml",
 		errMsg:   `"not$allowed" is not a valid label name`,
-	}, {
+	},
+	{
 		filename: "labelname2.bad.yml",
 		errMsg:   `"not:allowed" is not a valid label name`,
-	}, {
+	},
+	{
 		filename: "labelvalue.bad.yml",
 		errMsg:   `"\xff" is not a valid label value`,
-	}, {
+	},
+	{
 		filename: "regex.bad.yml",
 		errMsg:   "error parsing regexp",
-	}, {
+	},
+	{
 		filename: "modulus_missing.bad.yml",
 		errMsg:   "relabel configuration for hashmod requires non-zero modulus",
-	}, {
+	},
+	{
 		filename: "labelkeep.bad.yml",
 		errMsg:   "labelkeep action requires only 'regex', and no other fields",
-	}, {
+	},
+	{
 		filename: "labelkeep2.bad.yml",
 		errMsg:   "labelkeep action requires only 'regex', and no other fields",
-	}, {
+	},
+	{
 		filename: "labelkeep3.bad.yml",
 		errMsg:   "labelkeep action requires only 'regex', and no other fields",
-	}, {
+	},
+	{
 		filename: "labelkeep4.bad.yml",
 		errMsg:   "labelkeep action requires only 'regex', and no other fields",
-	}, {
+	},
+	{
 		filename: "labelkeep5.bad.yml",
 		errMsg:   "labelkeep action requires only 'regex', and no other fields",
-	}, {
+	},
+	{
 		filename: "labeldrop.bad.yml",
 		errMsg:   "labeldrop action requires only 'regex', and no other fields",
-	}, {
+	},
+	{
 		filename: "labeldrop2.bad.yml",
 		errMsg:   "labeldrop action requires only 'regex', and no other fields",
-	}, {
+	},
+	{
 		filename: "labeldrop3.bad.yml",
 		errMsg:   "labeldrop action requires only 'regex', and no other fields",
-	}, {
+	},
+	{
 		filename: "labeldrop4.bad.yml",
 		errMsg:   "labeldrop action requires only 'regex', and no other fields",
-	}, {
+	},
+	{
 		filename: "labeldrop5.bad.yml",
 		errMsg:   "labeldrop action requires only 'regex', and no other fields",
-	}, {
+	},
+	{
 		filename: "labelmap.bad.yml",
 		errMsg:   "\"l-$1\" is invalid 'replacement' for labelmap action",
-	}, {
+	},
+	{
 		filename: "rules.bad.yml",
 		errMsg:   "invalid rule file path",
-	}, {
+	},
+	{
 		filename: "unknown_attr.bad.yml",
 		errMsg:   "field consult_sd_configs not found in type",
-	}, {
+	},
+	{
 		filename: "bearertoken.bad.yml",
 		errMsg:   "at most one of bearer_token & bearer_token_file must be configured",
-	}, {
+	},
+	{
 		filename: "bearertoken_basicauth.bad.yml",
 		errMsg:   "at most one of basic_auth, oauth2, bearer_token & bearer_token_file must be configured",
-	}, {
+	},
+	{
 		filename: "kubernetes_http_config_without_api_server.bad.yml",
 		errMsg:   "to use custom HTTP client configuration please provide the 'api_server' URL explicitly",
-	}, {
+	},
+	{
 		filename: "kubernetes_kubeconfig_with_apiserver.bad.yml",
 		errMsg:   "cannot use 'kubeconfig_file' and 'api_server' simultaneously",
-	}, {
+	},
+	{
 		filename: "kubernetes_kubeconfig_with_http_config.bad.yml",
 		errMsg:   "cannot use a custom HTTP client configuration together with 'kubeconfig_file'",
 	},
 	{
 		filename: "kubernetes_bearertoken.bad.yml",
 		errMsg:   "at most one of bearer_token & bearer_token_file must be configured",
-	}, {
+	},
+	{
 		filename: "kubernetes_role.bad.yml",
 		errMsg:   "role",
-	}, {
+	},
+	{
 		filename: "kubernetes_selectors_endpoints.bad.yml",
 		errMsg:   "endpoints role supports only pod, service, endpoints selectors",
-	}, {
+	},
+	{
 		filename: "kubernetes_selectors_ingress.bad.yml",
 		errMsg:   "ingress role supports only ingress selectors",
-	}, {
+	},
+	{
 		filename: "kubernetes_selectors_node.bad.yml",
 		errMsg:   "node role supports only node selectors",
-	}, {
+	},
+	{
 		filename: "kubernetes_selectors_pod.bad.yml",
 		errMsg:   "pod role supports only pod selectors",
-	}, {
+	},
+	{
 		filename: "kubernetes_selectors_service.bad.yml",
 		errMsg:   "service role supports only service selectors",
-	}, {
+	},
+	{
 		filename: "kubernetes_namespace_discovery.bad.yml",
 		errMsg:   "field foo not found in type kubernetes.plain",
-	}, {
+	},
+	{
 		filename: "kubernetes_selectors_duplicated_role.bad.yml",
 		errMsg:   "duplicated selector role: pod",
-	}, {
+	},
+	{
 		filename: "kubernetes_selectors_incorrect_selector.bad.yml",
 		errMsg:   "invalid selector: 'metadata.status-Running'; can't understand 'metadata.status-Running'",
-	}, {
+	},
+	{
 		filename: "kubernetes_bearertoken_basicauth.bad.yml",
 		errMsg:   "at most one of basic_auth, oauth2, bearer_token & bearer_token_file must be configured",
-	}, {
+	},
+	{
 		filename: "kubernetes_authorization_basicauth.bad.yml",
 		errMsg:   "at most one of basic_auth, oauth2 & authorization must be configured",
-	}, {
+	},
+	{
 		filename: "marathon_no_servers.bad.yml",
 		errMsg:   "marathon_sd: must contain at least one Marathon server",
-	}, {
+	},
+	{
 		filename: "marathon_authtoken_authtokenfile.bad.yml",
 		errMsg:   "marathon_sd: at most one of auth_token & auth_token_file must be configured",
-	}, {
+	},
+	{
 		filename: "marathon_authtoken_basicauth.bad.yml",
 		errMsg:   "marathon_sd: at most one of basic_auth, auth_token & auth_token_file must be configured",
-	}, {
+	},
+	{
 		filename: "marathon_authtoken_bearertoken.bad.yml",
 		errMsg:   "marathon_sd: at most one of bearer_token, bearer_token_file, auth_token & auth_token_file must be configured",
-	}, {
+	},
+	{
 		filename: "marathon_authtoken_authorization.bad.yml",
 		errMsg:   "marathon_sd: at most one of auth_token, auth_token_file & authorization must be configured",
-	}, {
+	},
+	{
 		filename: "openstack_role.bad.yml",
 		errMsg:   "unknown OpenStack SD role",
-	}, {
+	},
+	{
 		filename: "openstack_availability.bad.yml",
 		errMsg:   "unknown availability invalid, must be one of admin, internal or public",
-	}, {
+	},
+	{
 		filename: "url_in_targetgroup.bad.yml",
 		errMsg:   "\"http://bad\" is not a valid hostname",
-	}, {
+	},
+	{
 		filename: "target_label_missing.bad.yml",
 		errMsg:   "relabel configuration for replace action requires 'target_label' value",
-	}, {
+	},
+	{
 		filename: "target_label_hashmod_missing.bad.yml",
 		errMsg:   "relabel configuration for hashmod action requires 'target_label' value",
-	}, {
+	},
+	{
 		filename: "unknown_global_attr.bad.yml",
 		errMsg:   "field nonexistent_field not found in type config.plain",
-	}, {
+	},
+	{
 		filename: "remote_read_url_missing.bad.yml",
 		errMsg:   `url for remote_read is empty`,
-	}, {
+	},
+	{
 		filename: "remote_write_header.bad.yml",
 		errMsg:   `x-prometheus-remote-write-version is a reserved header. It must not be changed`,
-	}, {
+	},
+	{
 		filename: "remote_read_header.bad.yml",
 		errMsg:   `x-prometheus-remote-write-version is a reserved header. It must not be changed`,
-	}, {
+	},
+	{
 		filename: "remote_write_authorization_header.bad.yml",
 		errMsg:   `authorization header must be changed via the basic_auth, authorization, oauth2, or sigv4 parameter`,
-	}, {
+	},
+	{
 		filename: "remote_write_url_missing.bad.yml",
 		errMsg:   `url for remote_write is empty`,
-	}, {
+	},
+	{
 		filename: "remote_write_dup.bad.yml",
 		errMsg:   `found multiple remote write configs with job name "queue1"`,
-	}, {
+	},
+	{
 		filename: "remote_read_dup.bad.yml",
 		errMsg:   `found multiple remote read configs with job name "queue1"`,
 	},
@@ -1262,6 +1373,22 @@ var expectedErrors = []struct {
 		filename: "empty_static_config.bad.yml",
 		errMsg:   "empty or null section in static_configs",
 	},
+	{
+		filename: "puppetdb_no_query.bad.yml",
+		errMsg:   "query missing",
+	},
+	{
+		filename: "puppetdb_no_url.bad.yml",
+		errMsg:   "URL is missing",
+	},
+	{
+		filename: "puppetdb_bad_url.bad.yml",
+		errMsg:   "host is missing in URL",
+	},
+	{
+		filename: "puppetdb_no_scheme.bad.yml",
+		errMsg:   "URL scheme must be 'http' or 'https'",
+	},
 	{
 		filename: "hetzner_role.bad.yml",
 		errMsg:   "unknown role",
@@ -1306,11 +1433,15 @@ var expectedErrors = []struct {
 		filename: "empty_scrape_config_action.bad.yml",
 		errMsg:   "relabel action cannot be empty",
 	},
+	{
+		filename: "uyuni_no_server.bad.yml",
+		errMsg:   "Uyuni SD configuration requires server host",
+	},
 }
 
 func TestBadConfigs(t *testing.T) {
 	for _, ee := range expectedErrors {
-		_, err := LoadFile("testdata/"+ee.filename, false, log.NewNopLogger())
+		_, err := LoadFile("testdata/"+ee.filename, false, false, log.NewNopLogger())
 		require.Error(t, err, "%s", ee.filename)
 		require.Contains(t, err.Error(), ee.errMsg,
 			"Expected error for %s to contain %q but got: %s", ee.filename, ee.errMsg, err)
@@ -1344,20 +1475,20 @@ func TestExpandExternalLabels(t *testing.T) {
 	// Cleanup ant TEST env variable that could exist on the system.
 	os.Setenv("TEST", "")
 
-	c, err := LoadFile("testdata/external_labels.good.yml", false, log.NewNopLogger())
+	c, err := LoadFile("testdata/external_labels.good.yml", false, false, log.NewNopLogger())
 	require.NoError(t, err)
 	require.Equal(t, labels.Label{Name: "bar", Value: "foo"}, c.GlobalConfig.ExternalLabels[0])
 	require.Equal(t, labels.Label{Name: "baz", Value: "foo${TEST}bar"}, c.GlobalConfig.ExternalLabels[1])
 	require.Equal(t, labels.Label{Name: "foo", Value: "${TEST}"}, c.GlobalConfig.ExternalLabels[2])
 
-	c, err = LoadFile("testdata/external_labels.good.yml", true, log.NewNopLogger())
+	c, err = LoadFile("testdata/external_labels.good.yml", false, true, log.NewNopLogger())
 	require.NoError(t, err)
 	require.Equal(t, labels.Label{Name: "bar", Value: "foo"}, c.GlobalConfig.ExternalLabels[0])
 	require.Equal(t, labels.Label{Name: "baz", Value: "foobar"}, c.GlobalConfig.ExternalLabels[1])
 	require.Equal(t, labels.Label{Name: "foo", Value: ""}, c.GlobalConfig.ExternalLabels[2])
 
 	os.Setenv("TEST", "TestValue")
-	c, err = LoadFile("testdata/external_labels.good.yml", true, log.NewNopLogger())
+	c, err = LoadFile("testdata/external_labels.good.yml", false, true, log.NewNopLogger())
 	require.NoError(t, err)
 	require.Equal(t, labels.Label{Name: "bar", Value: "foo"}, c.GlobalConfig.ExternalLabels[0])
 	require.Equal(t, labels.Label{Name: "baz", Value: "fooTestValuebar"}, c.GlobalConfig.ExternalLabels[1])
