@@ -14,18 +14,20 @@
 package kubernetes
 
 import (
+	"context"
 	"testing"
 
 	"github.com/prometheus/common/model"
-	"github.com/prometheus/prometheus/discovery/targetgroup"
 	v1 "k8s.io/api/core/v1"
 	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
 	"k8s.io/apimachinery/pkg/runtime"
 	"k8s.io/apimachinery/pkg/types"
+
+	"github.com/prometheus/prometheus/discovery/targetgroup"
 )
 
 func makeEndpoints() *v1.Endpoints {
-	var nodeName = "foobar"
+	nodeName := "foobar"
 	return &v1.Endpoints{
 		ObjectMeta: metav1.ObjectMeta{
 			Name:      "testendpoints",
@@ -78,7 +80,7 @@ func TestEndpointsDiscoveryBeforeRun(t *testing.T) {
 		discovery: n,
 		beforeRun: func() {
 			obj := makeEndpoints()
-			c.CoreV1().Endpoints(obj.Namespace).Create(obj)
+			c.CoreV1().Endpoints(obj.Namespace).Create(context.Background(), obj, metav1.CreateOptions{})
 		},
 		expectedMaxItems: 1,
 		expectedRes: map[string]*targetgroup.Group{
@@ -184,7 +186,7 @@ func TestEndpointsDiscoveryAdd(t *testing.T) {
 					},
 				},
 			}
-			c.CoreV1().Endpoints(obj.Namespace).Create(obj)
+			c.CoreV1().Endpoints(obj.Namespace).Create(context.Background(), obj, metav1.CreateOptions{})
 		},
 		expectedMaxItems: 1,
 		expectedRes: map[string]*targetgroup.Group{
@@ -241,7 +243,7 @@ func TestEndpointsDiscoveryDelete(t *testing.T) {
 		discovery: n,
 		afterStart: func() {
 			obj := makeEndpoints()
-			c.CoreV1().Endpoints(obj.Namespace).Delete(obj.Name, &metav1.DeleteOptions{})
+			c.CoreV1().Endpoints(obj.Namespace).Delete(context.Background(), obj.Name, metav1.DeleteOptions{})
 		},
 		expectedMaxItems: 2,
 		expectedRes: map[string]*targetgroup.Group{
@@ -294,7 +296,7 @@ func TestEndpointsDiscoveryUpdate(t *testing.T) {
 					},
 				},
 			}
-			c.CoreV1().Endpoints(obj.Namespace).Update(obj)
+			c.CoreV1().Endpoints(obj.Namespace).Update(context.Background(), obj, metav1.UpdateOptions{})
 		},
 		expectedMaxItems: 2,
 		expectedRes: map[string]*targetgroup.Group{
@@ -336,7 +338,7 @@ func TestEndpointsDiscoveryEmptySubsets(t *testing.T) {
 				},
 				Subsets: []v1.EndpointSubset{},
 			}
-			c.CoreV1().Endpoints(obj.Namespace).Update(obj)
+			c.CoreV1().Endpoints(obj.Namespace).Update(context.Background(), obj, metav1.UpdateOptions{})
 		},
 		expectedMaxItems: 2,
 		expectedRes: map[string]*targetgroup.Group{
@@ -366,7 +368,7 @@ func TestEndpointsDiscoveryWithService(t *testing.T) {
 					},
 				},
 			}
-			c.CoreV1().Services(obj.Namespace).Create(obj)
+			c.CoreV1().Services(obj.Namespace).Create(context.Background(), obj, metav1.CreateOptions{})
 		},
 		expectedMaxItems: 1,
 		expectedRes: map[string]*targetgroup.Group{
@@ -421,7 +423,7 @@ func TestEndpointsDiscoveryWithServiceUpdate(t *testing.T) {
 					},
 				},
 			}
-			c.CoreV1().Services(obj.Namespace).Create(obj)
+			c.CoreV1().Services(obj.Namespace).Create(context.Background(), obj, metav1.CreateOptions{})
 		},
 		afterStart: func() {
 			obj := &v1.Service{
@@ -434,7 +436,7 @@ func TestEndpointsDiscoveryWithServiceUpdate(t *testing.T) {
 					},
 				},
 			}
-			c.CoreV1().Services(obj.Namespace).Update(obj)
+			c.CoreV1().Services(obj.Namespace).Update(context.Background(), obj, metav1.UpdateOptions{})
 		},
 		expectedMaxItems: 2,
 		expectedRes: map[string]*targetgroup.Group{
