@@ -1,5 +1,5 @@
 import React, { FC, useState } from 'react';
-import { Link } from '@reach/router';
+import { Link } from 'react-router-dom';
 import {
   Collapse,
   Navbar,
@@ -17,17 +17,18 @@ import { ThemeToggle } from './Theme';
 
 interface NavbarProps {
   consolesLink: string | null;
+  agentMode: boolean;
 }
 
-const Navigation: FC<NavbarProps> = ({ consolesLink }) => {
+const Navigation: FC<NavbarProps> = ({ consolesLink, agentMode }) => {
   const [isOpen, setIsOpen] = useState(false);
   const toggle = () => setIsOpen(!isOpen);
   const pathPrefix = usePathPrefix();
   return (
     <Navbar className="mb-3" dark color="dark" expand="md" fixed="top">
       <NavbarToggler onClick={toggle} className="mr-2" />
-      <Link className="pt-0 navbar-brand" to={`${pathPrefix}/graph`}>
-        Prometheus
+      <Link className="pt-0 navbar-brand" to={agentMode ? '/agent' : '/graph'}>
+        Prometheus{agentMode && ' Agent'}
       </Link>
       <Collapse isOpen={isOpen} navbar style={{ justifyContent: 'space-between' }}>
         <Nav className="ml-0" navbar>
@@ -36,40 +37,48 @@ const Navigation: FC<NavbarProps> = ({ consolesLink }) => {
               <NavLink href={consolesLink}>Consoles</NavLink>
             </NavItem>
           )}
-          <NavItem>
-            <NavLink tag={Link} to={`${pathPrefix}/alerts`}>
-              Alerts
-            </NavLink>
-          </NavItem>
-          <NavItem>
-            <NavLink tag={Link} to={`${pathPrefix}/graph`}>
-              Graph
-            </NavLink>
-          </NavItem>
+          {!agentMode && (
+            <>
+              <NavItem>
+                <NavLink tag={Link} to="/alerts">
+                  Alerts
+                </NavLink>
+              </NavItem>
+              <NavItem>
+                <NavLink tag={Link} to="/graph">
+                  Graph
+                </NavLink>
+              </NavItem>
+            </>
+          )}
           <UncontrolledDropdown nav inNavbar>
             <DropdownToggle nav caret>
               Status
             </DropdownToggle>
             <DropdownMenu>
-              <DropdownItem tag={Link} to={`${pathPrefix}/status`}>
+              <DropdownItem tag={Link} to="/status">
                 Runtime & Build Information
               </DropdownItem>
-              <DropdownItem tag={Link} to={`${pathPrefix}/tsdb-status`}>
-                TSDB Status
-              </DropdownItem>
-              <DropdownItem tag={Link} to={`${pathPrefix}/flags`}>
+              {!agentMode && (
+                <DropdownItem tag={Link} to="/tsdb-status">
+                  TSDB Status
+                </DropdownItem>
+              )}
+              <DropdownItem tag={Link} to="/flags">
                 Command-Line Flags
               </DropdownItem>
-              <DropdownItem tag={Link} to={`${pathPrefix}/config`}>
+              <DropdownItem tag={Link} to="/config">
                 Configuration
               </DropdownItem>
-              <DropdownItem tag={Link} to={`${pathPrefix}/rules`}>
-                Rules
-              </DropdownItem>
-              <DropdownItem tag={Link} to={`${pathPrefix}/targets`}>
+              {!agentMode && (
+                <DropdownItem tag={Link} to="/rules">
+                  Rules
+                </DropdownItem>
+              )}
+              <DropdownItem tag={Link} to="/targets">
                 Targets
               </DropdownItem>
-              <DropdownItem tag={Link} to={`${pathPrefix}/service-discovery`}>
+              <DropdownItem tag={Link} to="/service-discovery">
                 Service Discovery
               </DropdownItem>
             </DropdownMenu>
@@ -77,9 +86,11 @@ const Navigation: FC<NavbarProps> = ({ consolesLink }) => {
           <NavItem>
             <NavLink href="https://prometheus.io/docs/prometheus/latest/getting_started/">Help</NavLink>
           </NavItem>
-          <NavItem>
-            <NavLink href={`${pathPrefix}/classic/graph${window.location.search}`}>Classic UI</NavLink>
-          </NavItem>
+          {!agentMode && (
+            <NavItem>
+              <NavLink href={`${pathPrefix}/classic/graph${window.location.search}`}>Classic UI</NavLink>
+            </NavItem>
+          )}
         </Nav>
       </Collapse>
       <ThemeToggle />
