@@ -32,6 +32,7 @@ import (
 	"regexp"
 	"runtime"
 	"sort"
+	"strconv"
 	"strings"
 	"sync"
 	template_text "text/template"
@@ -70,18 +71,27 @@ import (
 
 // Paths that are handled by the React / Reach router that should all be served the main React app's index.html.
 var reactRouterPaths = []string{
-	"/alerts",
 	"/config",
 	"/flags",
-	"/graph",
-	"/rules",
 	"/service-discovery",
 	"/status",
 	"/targets",
-	"/tsdb-status",
 	"/starting",
 }
 
+// Paths that are handled by the React router when the Agent mode is set.
+var reactRouterAgentPaths = []string{
+	"/agent",
+}
+
+// Paths that are handled by the React router when the Agent mode is not set.
+var reactRouterServerPaths = []string{
+	"/alerts",
+	"/graph",
+	"/rules",
+	"/tsdb-status",
+}
+
 // withStackTrace logs the stack trace in case the request panics. The function
 // will re-raise the error which will then be handled by the net/http package.
 // It is needed because the go-kit log package doesn't manage properly the
@@ -248,6 +258,7 @@ type Options struct {
 	RemoteReadConcurrencyLimit int
 	RemoteReadBytesInFrame     int
 	RemoteWriteReceiver        bool
+	IsAgent                    bool
 
 	Gatherer   prometheus.Gatherer
 	Registerer prometheus.Registerer
@@ -328,6 +339,7 @@ func New(logger log.Logger, o *Options) *Handler {
 		h.options.RemoteReadSampleLimit,
 		h.options.RemoteReadConcurrencyLimit,
 		h.options.RemoteReadBytesInFrame,
+		h.options.IsAgent,
 		h.options.CORSOrigin,
 		h.runtimeInfo,
 		h.versionInfo,
@@ -343,10 +355,15 @@ func New(logger log.Logger, o *Options) *Handler {
 		router = router.WithPrefix(o.RoutePrefix)
 	}
 
+	homePage := "/graph"
+	if o.IsAgent {
+		homePage = "/agent"
+	}
+
 	readyf := h.testReady
 
 	router.Get("/", func(w http.ResponseWriter, r *http.Request) {
-		http.Redirect(w, r, path.Join(o.ExternalURL.Path, "/graph"), http.StatusFound)
+		http.Redirect(w, r, path.Join(o.ExternalURL.Path, homePage), http.StatusFound)
 	})
 	router.Get("/classic/", func(w http.ResponseWriter, r *http.Request) {
 		http.Redirect(w, r, path.Join(o.ExternalURL.Path, "/classic/graph"), http.StatusFound)
@@ -397,6 +414,7 @@ func New(logger log.Logger, o *Options) *Handler {
 			fmt.Fprintf(w, "Error opening React index.html: %v", err)
 			return
 		}
+		defer func() { _ = f.Close() }()
 		idx, err := ioutil.ReadAll(f)
 		if err != nil {
 			w.WriteHeader(http.StatusInternalServerError)
@@ -405,6 +423,7 @@ func New(logger log.Logger, o *Options) *Handler {
 		}
 		replacedIdx := bytes.ReplaceAll(idx, []byte("CONSOLES_LINK_PLACEHOLDER"), []byte(h.consolesPath()))
 		replacedIdx = bytes.ReplaceAll(replacedIdx, []byte("TITLE_PLACEHOLDER"), []byte(h.options.PageTitle))
+		replacedIdx = bytes.ReplaceAll(replacedIdx, []byte("AGENT_MODE_PLACEHOLDER"), []byte(strconv.FormatBool(h.options.IsAgent)))
 		w.Write(replacedIdx)
 	}
 
@@ -413,6 +432,16 @@ func New(logger log.Logger, o *Options) *Handler {
 		router.Get(p, serveReactApp)
 	}
 
+	if h.options.IsAgent {
+		for _, p := range reactRouterAgentPaths {
+			router.Get(p, serveReactApp)
+		}
+	} else {
+		for _, p := range reactRouterServerPaths {
+			router.Get(p, serveReactApp)
+		}
+	}
+
 	// The favicon and manifest are bundled as part of the React app, but we want to serve
 	// them on the root.
 	for _, p := range []string{"/favicon.ico", "/manifest.json"} {
@@ -589,7 +618,7 @@ func (h *Handler) Run(ctx context.Context, listener net.Listener, webConfig stri
 		ReadTimeout: h.options.ReadTimeout,
 	}
 
-	errCh := make(chan error)
+	errCh := make(chan error, 1)
 	go func() {
 		errCh <- toolkit_web.Serve(listener, httpSrv, webConfig, h.logger)
 	}()
@@ -604,7 +633,6 @@ func (h *Handler) Run(ctx context.Context, listener net.Listener, webConfig stri
 }
 
 func (h *Handler) alerts(w http.ResponseWriter, r *http.Request) {
-
 	var groups []*rules.Group
 	for _, group := range h.ruleManager.RuleGroups() {
 		if group.HasAlertingRules() {
@@ -709,6 +737,7 @@ func (h *Handler) consoles(w http.ResponseWriter, r *http.Request) {
 		h.now(),
 		template.QueryFunc(rules.EngineQueryFunc(h.queryEngine, h.storage)),
 		h.options.ExternalURL,
+		nil,
 	)
 	filenames, err := filepath.Glob(h.options.ConsoleLibrariesPath + "/*.lib")
 	if err != nil {
@@ -1113,6 +1142,7 @@ func (h *Handler) executeTemplate(w http.ResponseWriter, name string, data inter
 		h.now(),
 		template.QueryFunc(rules.EngineQueryFunc(h.queryEngine, h.storage)),
 		h.options.ExternalURL,
+		nil,
 	)
 	tmpl.Funcs(tmplFuncs(h.consolesPath(), h.options))
 
